/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <stdarg.h>
#include <ctype.h>
#include <qparse/qparse.h>

#include "window.h"

#include "action.h"

fgw_ctx_t fgw;
static char *action_str_buf = NULL;
static size_t action_str_buf_len = 0;
const char actd_signature[] = "action details signature";
const irc_from_t *action_null_from;

#define ACTION_SYNTAX (QPARSE_DOUBLE_QUOTE | QPARSE_SINGLE_QUOTE | QPARSE_MULTISEP | QPARSE_TERM_NEWLINE | QPARSE_TERM_SEMICOLON | QPARSE_SEP_COMMA | QPARSE_PAREN_FUNC)
#define USER_INPUT_SYNTAX  QPARSE_MULTISEP

void action_help_print(const char *str, ...)
{
	win_t *curr = idx2win(win_curr_idx);
	const char *line;
	va_list ap;

	win_append(curr, "");
	win_append(curr, "--- help begin ---");
	win_append(curr, str);

	va_start(ap, str);
	while((line = va_arg(ap, const char *)) != NULL)
		win_append(curr, line);
	va_end(ap);


	win_append(curr, "--- help end ---");
	win_append(curr, "");
}

void action_strarr(action_details_t *details, int argc, char **argv)
{
	fgw_arg_t fargv[ACTION_MAX_ARGC+1];
	int n;

	for(n = 1; n < argc; n++) {
		fargv[n].type = FGW_STR;
		fargv[n].val.str = argv[n];
	}

	if ((argv[0][0] != 'o') || (argv[0][1] != 'n') || (argv[0][2] != '_')) {
		fgw_func_t *f = fgw_func_lookup(&fgw, argv[0]);
		if (f == NULL) {
			win_t *curr = idx2win(win_curr_idx);
			win_append(curr, "*** Error: no such action:");
			win_append(curr, argv[0]);
			return;
		}
	}

	fgw_ucall_all(&fgw, details, argv[0], argc, fargv);
}

void action_vstr(const char *name, ...)
{
	action_details_t details;
	const char *argv[ACTION_MAX_ARGC] = {0};
	int argc = 0;
	va_list ap;

	argv[argc++] = name;
	va_start(ap, name);
	while((argv[argc++] = va_arg(ap, const char *)) != NULL)
		if (argc >= ACTION_MAX_ARGC-1)
			abort();
	argv[argc--] = NULL;
	va_end(ap);

	action_details_reset(&details);
	action_strarr(&details, argc, argv);
}


static void action_str_(const char *s, int syntax)
{
	size_t cons;
	action_details_t details;
	char *argv_static[ACTION_MAX_ARGC] = {0};
	char **argv = argv_static;
	unsigned int argv_allocated = ACTION_MAX_ARGC;

	action_details_reset(&details);

	while(*s != '\0') {
		int argc;

		while((*s == ';') || isspace(*s)) s++;

		if (*s == '\0') break;

		argc = qparse4(s, &argv, &argv_allocated, syntax | QPARSE_NO_ARGV_REALLOC, &cons, &action_str_buf, &action_str_buf_len, (const char **)&details.start, ACTION_MAX_ARGC);
		details.has_start = !(syntax & QPARSE_PAREN_FUNC);
		action_strarr(&details, argc, argv);
		qparse_free_strs(argc, &argv);
		s += cons;
	}
}

void action_str(const char *s)     { action_str_(s, ACTION_SYNTAX); }
void action_str_ui(const char *s)  { action_str_(s, USER_INPUT_SYNTAX); }

void action_load_file(const char *fn)
{
	FILE *f;
	char *s, line[8192];

	f = fopen(fn, "r");
	if (f == NULL)
		return;

	while((s = fgets(line, sizeof(line), f)) != NULL) {
		while(isspace(*s)) s++;
		if ((*s != '#') && (*s != '\0'))
			action_str(s);
	}
	fclose(f);
}

void act_error(const char *actname, const char *msg)
{
	win_t *curr = idx2win(win_curr_idx);
	win_append(curr, msg);
}


void action_init(void)
{
	fgw_init(&fgw, "fgwirc");
}

void action_uninit(void)
{
	fgw_uninit(&fgw);
	free(action_str_buf);
	action_str_buf_len = 0;
}

