#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "driver.h"

const char *tcurses_cap_search_sys[] = { PATH_INST_CAPS, NULL };
const char **tcurses_cap_search_user = NULL;

#define INT_ARG_MAX   (9999+1)
#define INT_ARG_WIDTH 4

static char *fmt_names[] = {
	"reset",
	"clear",
	"home",
	"gotoxy",
	"getxy",
	"GETXY",
	"incx",
	"setfg",
	"setbg",
	"seta",
	"setattr",
	"query_xy",
	"query_status",
	NULL
};

static char *strclone(const char *s)
{
	int l;
	char *o;
	l = strlen(s)+1;
	o = malloc(l);
	memcpy(o, s, l);
	return o;
}

int load_color(tcurses_drv_color_t **list, char *val)
{
	char *shex, *sval, *sattr, *sname;
	tcurses_drv_color_t *c;
#warning TODO
	return 0;
}

static int load_key(tcurses_drv_t *drv, const char *key, char *val)
{
	int fmt_idx;
	char **n;

	if (strcmp(key, "escape") == 0) {
		char *end;
		int v;
		v = strtol(val, &end, 0);
		if (*end != '\0')
			return -1;
		drv->esc = v;
		return 0;
	}

	if (strcmp(key, "fgcolor") == 0)
		return load_color(&(drv->fg_colors), val);

	if (strcmp(key, "bgcolor") == 0)
		return load_color(&(drv->fg_colors), val);

	for(fmt_idx = 0, n = fmt_names; *n != NULL; n++, fmt_idx++) {
		if (strcmp(*n, key) == 0) {
			const char *s;
			int len;
			drv->fmt[fmt_idx] = strclone(val);
			for(len = 0, s = val; *s != '\0'; s++) {
				if (*s == '%') {
					s++;
					switch(*s) {
						case '\0':
							s--;
						case 'e':
						case '%':
							len++;
							break;
						default:
							/* integer argument */
							len += INT_ARG_WIDTH;
					}
				}
				else
					len++;
			}
			if (len > drv->bufflen)
				drv->bufflen = len;
			return 0;
		}
	}
	return 0;
}

static FILE *open_caps(const char **search, const char *term_name, char *fn, long max_fn)
{
	long tnl = strlen(term_name), maxl = max_fn - 12;
	FILE *f;

	while(*search != NULL) {
		if (strlen(*search) + tnl < maxl) {
			sprintf(fn, "%s/%s.tcurses", *search, term_name);
			f = fopen(fn, "r");
			if (f != NULL)
				return f;
		}
		search++;
	}
	return NULL;
}

tcurses_drv_t *tcurses_drv_init(const char *term_name)
{
	char fn[8192];
	tcurses_drv_t *drv;
	FILE *f;
	char *line, *key, *val, *end;

	/* first search optional user provided paths */
	if (tcurses_cap_search_user != NULL)
		f = open_caps(tcurses_cap_search_user, term_name, fn, sizeof(fn));

	/* if that didn't work, search sys installed paths */
	if (f == NULL)
		f = open_caps(tcurses_cap_search_sys, term_name, fn, sizeof(fn));

	if (f == NULL)
		return NULL;

	drv = calloc(sizeof(tcurses_drv_t), 1);
	drv->esc = -1;
	while(!(feof(f))) {
		line = fn;
		*line = '\0';
		fgets(line, sizeof(fn), f);

		/* trim leading ws */
		while(isspace(*line)) line++;

		/* ignore empty lines and comments */
		if ((*line == '\0') || (*line == '\n') || (*line == '\r') || (*line == '#'))
			continue;

		/* trim \r\n at the end */
		end = strpbrk(line, "\r\n");
		if (end != NULL)
			*end = '\0';

		/* parse key value pairs */
		key = line;
		val = strpbrk(key, " \t");
		if (val != NULL) {
			*val = '\0';
			val++;
			while(isspace(*val)) val++;
			if (load_key(drv, key, val) != 0)
				goto error;
		}
	}

	fclose(f);
	return drv;
error:
	fclose(f);
	tcurses_drv_uninit(drv);
	return NULL;
}

void tcurses_drv_uninit(tcurses_drv_t *drv)
{
	int n;

	for(n = 0; n < FMT_num_fmts; n++)
		if (drv->fmt[n] != NULL)
			free(drv->fmt[n]);

	free(drv);
}

#define append(c) \
	do { \
		*s = (c); \
		s++; \
	} while(0)

int render(tcurses_drv_t *drv, char *buff, int fmt_idx, char *names, int *vals)
{
	char *f, *s, *name;
	int n, found;
	if (drv->fmt[fmt_idx] == NULL)
		return -1;
	for(s = buff, f = drv->fmt[fmt_idx]; *f != '\0'; f++) {
		if (*f == '%') {
			f++;
			switch(*f) {
				case '\0':
					append('%');
					goto end;
				case 'e':
					append(drv->esc);
					break;
				case '%':
					append('%');
					break;
				default:
					found = 0;
					for(n = 0, name = names; *name != '\0'; n++,name++) {
						if (*name == *f) {
							int len;
							if (vals[n] > INT_ARG_MAX)
								vals[n] = INT_ARG_MAX;
							len = sprintf(s, "%d", vals[n]);
							if (len < 0)
								return -1;
							s += len;
							found = 1;
							break;
						}
					}
					if (!found)
						return -1;
			}
		}
		else
			append(*f);
	}

	end:;
	append('\0');
	return s - buff - 1;
}
#undef append

#define noarg "\0", NULL

int tcurses_drv_reset(tcurses_drv_t *drv, char *buff)
{
	return render(drv, buff, FMT_RESET, noarg);
}

int tcurses_drv_clear(tcurses_drv_t *drv, char *buff)
{
	return render(drv, buff, FMT_CLEAR, noarg);
}

int tcurses_drv_home(tcurses_drv_t *drv, char *buff)
{
	return render(drv, buff, FMT_HOME, noarg);
}


int tcurses_drv_gotoxy(tcurses_drv_t *drv, char *buff, int col, int row)
{
	int args[4];
	args[0] = col;
	args[1] = row;
	args[2] = col+1;
	args[3] = row+1;
	return render(drv, buff, FMT_GOTOXY, "CRcr", args);
}

int tcurses_drv_incx(tcurses_drv_t *drv, char *buff, int col)
{
	int args[1];
	args[0] = col;
	return render(drv, buff, FMT_INCX, "C", args);
}


int tcurses_drv_setfg(tcurses_drv_t *drv, char *buff, char fg)
{
	int args[1];
	args[0] = fg;
	return render(drv, buff, FMT_SETFG, "F", args);
}

int tcurses_drv_setbg(tcurses_drv_t *drv, char *buff, char bg)
{
	int args[1];
	args[0] = bg;
	return render(drv, buff, FMT_SETBG, "B", args);
}

int tcurses_drv_seta(tcurses_drv_t *drv, char *buff, char a)
{
	int args[1];
	args[0] = a;
	return render(drv, buff, FMT_SETA, "A", args);
}


int tcurses_drv_setattr(tcurses_drv_t *drv, char *buff, char fg, char bg, char a)
{
	int args[3];
	args[0] = fg;
	args[1] = bg;
	args[2] = a;
	return render(drv, buff, FMT_SETATTR, "FBA", args);
}

/*** query and response processing ***/

int tcurses_drv_getxy(tcurses_drv_t *drv, char *buff)
{
	return render(drv, buff, FMT_GETXY_REQ, noarg);
}

int tcurses_drv_query_status(tcurses_drv_t *drv, char *buff);

int tcurses_drv_getch(tcurses_drv_t *drv, int ch)
{
	return ch;
}
