#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libporty/net/network.h>
#include "tcurses.h"
#include "driver.h"

enum {
	ST_SEARCHING_SX,
	ST_SEARCHING_SY,
	ST_IDLE
};

static void invalidate_attributes(tcurses_term_t *term)
{
	term->fg = -1;
	term->bg = -1;
	term->attr = -1;
	term->pfg = -1;
	term->pbg = -1;
	term->pattr = -1;
}

static void clear_or_reset(tcurses_term_t *term)
{
	int len;
	char buff[256];
	tcurses_drv_t *drv = (tcurses_drv_t *)term->driver;

	if ((len = tcurses_drv_clear(drv, buff)) > 0)
		P_net_write(term->fdo, buff, len);
	else {
		/* even worse, we have to reset the terminal as we can't just clear it */
		if ((len = tcurses_drv_reset(drv, buff)) > 0)
			P_net_write(term->fdo, buff, len);
		/* do not assume attributes */
		invalidate_attributes(term);
	}
}

void tcurses_clear(tcurses_term_t *term)
{
	clear_or_reset(term);
	memset(term->current, 0, sizeof(tcurses_cell_t) * term->sx * term->sy);
	memset(term->last, 0, sizeof(tcurses_cell_t) * term->sx * term->sy);
}

static void tcurses_clear_last(tcurses_term_t *term)
{
	memset(term->last, 0, sizeof(tcurses_cell_t) * term->sx * term->sy);
}

static void alloc_term(tcurses_term_t *term)
{
	term->current = calloc(sizeof(tcurses_cell_t), term->sx * term->sy);
	term->last    = calloc(sizeof(tcurses_cell_t), term->sx * term->sy);
	invalidate_attributes(term);
}

void tcurses_resize(tcurses_term_t *term, int sx, int sy)
{
	term->sx = sx;
	term->sy = sy;
	free(term->current);
	free(term->last);
	alloc_term(term);
	tcurses_clear(term);
}


tcurses_term_t *tcurses_init(int fdo, const char *term_type, int sx, int sy)
{
	tcurses_term_t *term;
	tcurses_drv_t *drv;
	drv = tcurses_drv_init(term_type);
	if (drv == NULL)
		return NULL;
	term = calloc(sizeof(tcurses_term_t), 1);
	term->driver = drv;
	term->fdo = fdo;
	if ((sx > 0) && (sy > 0)) {
		term->state = ST_IDLE;
		term->sx = sx;
		term->sy = sy;
		alloc_term(term);
	}
	else
		term->state = ST_SEARCHING_SX;
	return term;
}

void tcurses_uninit(tcurses_term_t *term)
{
	free(term);
}

int tcurses_inited(tcurses_term_t *term)
{
	return term->state >= ST_IDLE;
}

int tcurses_getch(tcurses_term_t *term, int ch)
{
	return ch;
}

#define cell_at_xy(term, which, x, y) ((term)->which[(x) + (y) * (term)->sx])

static void buff_flush(int fdo, char *buff, int *used)
{
	if (*used == 0)
		return;
	P_net_write(fdo, buff, *used);
	*used = 0;
}

static void buff_write_(int fdo, char *buff, int *used, int alloced, char *s, int len)
{
	if (len > alloced-(*used))
		buff_flush(fdo, buff, used);
	if (len >= alloced) {
		P_net_write(fdo, s, len);
		return;
	}
	memcpy(buff + (*used), s, len);
	*used += len;
}

#define buff_write(_fdo_, _str_, _len_) buff_write_((_fdo_), tmp, tmp_used, tmp_size, (_str_), (_len_))

static void check_home(tcurses_term_t *term, int *lx, int *ly, char *tmp, int *tmp_used, int tmp_size)
{
	tcurses_drv_t *drv = (tcurses_drv_t *)term->driver;
	int len;
	char buff[256];

	if (*lx < 0)
		*lx = 0;

	/* check if we need to go home */
	if ((*lx != 0) || (*ly != 0)) {
		/* we are already in home */
		return;
	}

	if ((len = tcurses_drv_home(drv, buff)) > 0)          /* try the dedicated command first */
		buff_write(term->fdo, buff, len);
	else if ((len = tcurses_drv_gotoxy(drv, buff, 0, 0)) > 0)    /* if that fails, gotoxy */
		buff_write(term->fdo, buff, len);
	else { /* if there's no other way, redraw the whole screen */
		clear_or_reset(term);
		tcurses_clear_last(term);
	}
}

/* compare curr to last and update as much as needed */
static void draw_curr(tcurses_term_t *term, tcurses_cell_t *curr, tcurses_cell_t *last, char *tmp, int *tmp_used, int tmp_size)
{
	tcurses_drv_t *drv = (tcurses_drv_t *)term->driver;
	int adiff = (curr->fg != term->pfg) + (curr->bg != term->pbg) + (curr->attr != term->pattr);
	int len;
	char buff[256];

	if (adiff > 0) {
		/* colors/attribute: if there are more than one to set, it's cheaper
		   to set all, if possible */
		if ((adiff > 1) && ((len = tcurses_drv_setattr(drv, buff, curr->fg, curr->bg, curr->attr)) > 0)) {
			buff_write(term->fdo, buff, len);
			term->pfg = curr->fg;
			term->pbg = curr->bg;
			term->pattr = curr->attr;
		}
		else {
			/* need to set one by one */
			if ((curr->fg != term->pfg) && ((len = tcurses_drv_setfg(drv, buff, curr->fg)) > 0)) {
				buff_write(term->fdo, buff, len);
				term->pfg = curr->fg;
			}
			if ((curr->bg != term->pbg) && ((len = tcurses_drv_setfg(drv, buff, curr->bg)) > 0)) {
				buff_write(term->fdo, buff, len);
				term->pbg = curr->bg;
			}
			if ((curr->attr != term->pattr) && ((len = tcurses_drv_seta(drv, buff, curr->attr)) > 0)) {
				buff_write(term->fdo, buff, len);
				term->pattr = curr->attr;
			}
		}
	}

	/* update the actual char */
	if (curr->c != last->c)
		buff_write(term->fdo, &(curr->c), 1);

	*last = *curr;
}

static void draw_until(tcurses_term_t *term, int fromx, int fromy, tcurses_cell_t *until_curr, char *tmp, int *tmp_used, int tmp_size)
{
	tcurses_cell_t *c, *l;
	c = &cell_at_xy(term, current, fromx, fromy);
	l = &cell_at_xy(term, last, fromx, fromy);
	for(; c < until_curr; c++, l++)
		draw_curr(term, c, l, tmp, tmp_used, tmp_size);
}


void tcurses_flush(tcurses_term_t *term)
{
	int x, y, lx, ly;
	tcurses_cell_t *c, *l;
	tcurses_drv_t *drv = (tcurses_drv_t *)term->driver;
	char buff[256];
	char tmp[1024];
	int len, tmp_used_ = 0, *tmp_used = &tmp_used_, tmp_size = sizeof(tmp);

	c = term->current;
	l = term->last;

	lx = -1;
	ly = 0;
	for(y = 0; y < term->sy; y++) {
		for(x = 0; x < term->sx; x++,c++,l++) {
			if ((c->c != l->c) || (c->fg != l->fg) || (c->bg != l->bg) || (c->attr != l->attr)) {
				if ((ly == y) && (lx == x-1)) {
					/* next character - check for home, but then just write it */
					if ((y == 0) && (x == 0)) {
						/* we need to go to home */
						check_home(term, &lx, &ly, tmp, tmp_used, tmp_size);
					}
				}
				else if ((ly == y) && ((len = tcurses_drv_incx(drv, buff, x - lx - 1)) > 0)) {
					/* skipped some chars, still the same row */
					buff_write(term->fdo, buff, len);
				}
				else if ((len = tcurses_drv_gotoxy(drv, buff, x, y)) > 0) {
					/* large jump */
					buff_write(term->fdo, buff, len);
				}
				else {
					/* all else has failed, have to print all characters in between */
					check_home(term, &lx, &ly, tmp, tmp_used, tmp_size);
					/* redraw only if there is more than zero of them */
					if ((x > lx+1) || (y != ly))
						draw_until(term, lx, ly, c, tmp, tmp_used, tmp_size);
				}

				draw_curr(term, c, l, tmp, tmp_used, tmp_size);

				/* update last drawn pos */
				lx = x;
				ly = y;
			}
		}
	}

	if ((len = tcurses_drv_gotoxy(drv, buff, term->pcx, term->pcy)) > 0)
		buff_write(term->fdo, buff, len);

	buff_flush(term->fdo, tmp, tmp_used);
}

static int limits(int v, int min, int max)
{
	if (v < min)
		return min;
	if (v > max)
		return max;
	return v;
}

void tcurses_cursor(tcurses_term_t *term, int col, int row)
{
	term->pcx = limits(col, 0, term->sx - 1);
	term->pcy = limits(row, 0, term->sy - 1);
}


void tcurses_gotoxy(tcurses_term_t *term, int col, int row)
{
	term->x = limits(col, 0, term->sx - 1);
	term->y = limits(row, 0, term->sy - 1);
}

void tcurses_setattr(tcurses_term_t *term, char fg, char bg, char a)
{
	term->fg = fg;
	term->bg = bg;
	term->attr = a;
}

static tcurses_cell_t *tcurses_write_chr(tcurses_term_t *term, tcurses_cell_t *c, int chr)
{
	c->fg   = term->fg;
	c->bg   = term->bg;
	c->attr = term->attr;
	c->c    = chr;
	c++;
	term->x++;
	if (term->x >= term->sx) {
		term->x = 0;
		term->y++;
		if (term->y >= term->sy)
			return c;
		c = &cell_at_xy(term, current, term->x, term->y);
	}
	return c;
}

void tcurses_write(tcurses_term_t *term, char *str)
{
	char *s;
	tcurses_cell_t *c;

	c = &cell_at_xy(term, current, term->x, term->y);
	for(s = str; *s != '\0'; s++)
		c = tcurses_write_chr(term, c, *s);
}


void tcurses_write_len(tcurses_term_t *term, char *str, int len)
{
	char *s;
	tcurses_cell_t *c;

	c = &cell_at_xy(term, current, term->x, term->y);
	for(s = str; len > 0; len--,s++)
		c = tcurses_write_chr(term, c, *s);
}

void tcurses_dump_to_file(tcurses_term_t *term, const char *fn)
{
	FILE *f;
	int x, y;
	tcurses_cell_t *c;

	f = fopen(fn, "a");
	fprintf(f, "### TCURSES DUMP ###\n");

	fprintf(f, "current char\n");
	c = term->current;
	for(y = 0; y < term->sy; y++) {
		for(x = 0; x < term->sx; x++,c++) {
			if (c->c == 0)
				fprintf(f, "\311");
			else
				fprintf(f, "%c", c->c);
		}
		fprintf(f, "\n");
	}

	fprintf(f, "current attr\n");
	c = term->current;
	for(y = 0; y < term->sy; y++) {
		for(x = 0; x < term->sx; x++,c++) {
			if (c->attr < 0)
				fprintf(f, "-");
			else
				fprintf(f, "%d", c->attr);
		}
		fprintf(f, "\n");
	}

	fclose(f);
}
